@@ -3,6 +3,7 @@
 import sys
 import time
 import threading
+import queue
 import traceback
 import atexit
 
@@ -27,6 +28,15 @@ global_channels = []
 t_start = time.time()
 
 
+def _print(s):
+    # sys.stdout.write(s)
+    print(s, end='')
+
+
+def _get_print_thread(target):
+    return threading.Thread(target=target, daemon=True)
+
+
 class Logger:
     def __init__(self, layout_leading_line='oneliner'):
         # FOR DEBUGGING
@@ -37,14 +47,11 @@ class Logger:
         self.set_refresh_rate(25)
         self.t_last_update = time.time() - self.refresh_interval
 
-        # Output string that will be printed to terminal during every update and lock object
-        self.outputs = []
-
         # Output event handling
-        self.lock_output = threading.Lock()
-        self.event_wait_interval = threading.Event()
+        # self.lock_output = threading.Lock()
+        # self.event_wait_interval = threading.Event()
         self.event_print_requested = threading.Event()
-        self.cond_print_requested = threading.Condition()
+        # self.cond_print_requested = threading.Condition()
 
         # Channels that can be registered, e.g. by different modules
         global global_default_channel
@@ -65,12 +72,46 @@ class Logger:
 
         # Activate ANSI escape codes if necessary
         activate_esc()
-        sys.stdout.write(COLOR_DEFAULT_FONT_FG + COLOR_DEFAULT_FONT_BG)
+        _print(COLOR_DEFAULT_FONT_FG + COLOR_DEFAULT_FONT_BG)
 
-        self.is_running = True
-        self.th_print = threading.Thread(target=self._on_print)
+        self._q = queue.Queue()
+        self._event_stop = threading.Event()
 
-    def _on_exception(self, exc_type, exc_value, exc_traceback):
+        self._th_print = threading.Thread(target=self._on_print, daemon=True)
+        self._th_print.start()
+
+    def is_running(self):
+        return not self._event_stop.is_set()
+
+    def on_exit(self):
+        """
+        Prints total runtime when Python exits
+        Note:
+            Does not make sense to make this a method of logger. Logger will be cleaned up first (?).
+            The Method would then still be executed but can not access attributes of logger anymore.
+            Thus, we need to use globals anyway.
+        """
+
+        # print('_on_exit 1')
+        self._stop_print_daemon()
+        # print('_on_exit 2')
+
+        # if exception is None:
+        t_elapsed = time.time() - t_start
+        str_timer_info = ('\n\nTotal runtime: {} ' + LIGHT_GRAY + '(since logbox import)' + RESET + '\n').format(
+            format_time(t_elapsed))
+        _print(str_timer_info)
+
+        # Explicitly log time to all channels
+        for channel in global_channels:
+            channel.on_log(str_timer_info + ' ' + format_time(time.time() - t_start), global_default_channel)
+
+        # Deactivate ANSI escape codes if necessary
+        # _print(RESET)
+        # time.sleep(0.1)
+        deactivate_esc()
+
+    def on_exception(self, exc_type, exc_value, exc_traceback):
         """
         Callback for exceptions
         Exceptions will be logged to each channel and then reraised.
@@ -83,7 +124,7 @@ class Logger:
 
         # Handle exception in case there is one
         if exc_type is not None:
-            # Note the exception for exit handling
+            # Note the exception for exit handling (global)
             exception = (exc_type, exc_value, exc_traceback)
 
             # Parse the traceback
@@ -98,8 +139,26 @@ class Logger:
             for channel in self.channels:
                 channel.on_error('\n\n' + str_traceback, self.default_channel)
 
+            # Wait until all outputs have been printed
+            self._stop_print_daemon()
+            # print('on_exception', exc_type)
+
             return self._reraise(exc_type, exc_value, exc_traceback)
 
+        # Wait until all outputs have been printed
+        # print('on_exception', exc_type)
+        self._stop_print_daemon()
+
+    # def __del__(self):
+    #     print('__del__')
+    #     # Wait until all outputs have been printed
+    #     self._wait_for_print_finished()
+
+    def _stop_print_daemon(self):
+        self._event_stop.set()
+        self._q.join()
+        self._th_print.join()
+
     def _print_time_since_start(self, *args):
         """ Print the total runtime since the logger has been started """
         t_elapsed = time.time() - t_start
@@ -111,13 +170,6 @@ class Logger:
         self.print(*args[:-1], str_time)
 
     def _reraise(self, exc_type, exc_value, exc_traceback):
-        # Wait until all outputs have been printed
-        while self.th_print.is_alive():
-            pass
-
-        # Add newlines
-        # self.print('\n')
-
         # Re-raise the exception
         return sys.__excepthook__(exc_type, exc_value, exc_traceback)
 
@@ -130,7 +182,6 @@ class Logger:
 
         # Can be given multiple arguments to be joined into one string (as Python's standard print-function)
         str_new = ' '.join([str(item) for item in str_new])
-        # str_new = ' '.join(np.asarray(str_new).astype('str'))  # --> Error with torch.Size
 
         # Get channel (in case name is given)
         channel = self._get_channel(channel)
@@ -145,17 +196,15 @@ class Logger:
         channel.on_log(str_new, self.default_channel)
 
         # Queue to terminal outputs
-        with self.lock_output:
-            self.outputs.append(str_new)
+        # with self.lock_output:
+        #     self.outputs.append(str_new)
+        self._q.put(str_new)
 
-        self.notify_print()
+        # self.notify_print()
 
     def notify_print(self):
         """ Serves as callback for progbar to request a print update """
         self.event_print_requested.set()
-        if not self.th_print.is_alive():
-            self.th_print = threading.Thread(target=self._on_print)
-            self.th_print.start()
 
     def _get_channel(self, channel):
         """ Chooses corresponding channel if specified by name """
@@ -209,38 +258,39 @@ class Logger:
         First check refresh timer, sleep if necessary and then print.
         """
 
-        while self.event_print_requested.is_set():
-            # Set flag for blocking further invocations during refresh interval
-            self.event_wait_interval.set()
-
+        while self.is_running():
             # Sleep until end of refresh interval
-            t_wait = max(0, self.refresh_interval - (time.time() - self.t_last_update))
+            t_wait = max(0.0, self.refresh_interval - (time.time() - self.t_last_update))
             time.sleep(t_wait)
 
-            # Assemble strings for leading line
-            leading_line_done, leading_line_active = self.leading_line.get_leading_line()
+            outputs = []
+            # try:
+            #     outputs.append(self._q.get(block=True, timeout=self.refresh_interval))
+            # except queue.Empty():
+            #     continue
+            while not self._q.empty():
+                outputs.append(self._q.get(block=False))
+                self._q.task_done()
 
-            # Lock output string
-            with self.lock_output:
-                self.event_print_requested.clear()
+            # TBD: do the following only if outputs are there or leading line requires to be printed again --> make notify_print useful again...
+
+            if len(outputs) > 0 or self.event_print_requested.is_set():
+                # Assemble strings for leading line
+                leading_line_done, leading_line_active = self.leading_line.get_leading_line()
 
                 # Update timer and reset flag
                 self.t_last_update = time.time()
-                self.event_wait_interval.clear()
 
                 # Remove previous leading line and add the current one
-                str_out = ERASE_LINE + CARRIAGE_RETURN + RESET + leading_line_done + ''.join(self.outputs) + leading_line_active
+                str_out = ERASE_LINE + CARRIAGE_RETURN + RESET + leading_line_done + ''.join(outputs) + leading_line_active
 
-                # Reset output queue
-                if len(self.outputs) > 0:
+                if len(outputs) > 0:
                     self.leading_line.is_headline_overwritten = True
                 else:
                     self.leading_line.is_headline_overwritten = False
-                self.outputs = []
 
-            # Print to terminal
-            sys.stdout.write(str_out)
-            # print(str_out, end='')
+                # Print to terminal
+                _print(str_out)
 
     def register(self, name, mode='sub', **params):
         """ Method that will be invoked by logger.register(...) to set up a new channel """
@@ -448,6 +498,7 @@ class LeadingLineTable:
         return headline, leading_line
 
 
+'''
 def _on_exit():
     """
     Prints total runtime when Python exits
@@ -456,27 +507,37 @@ def _on_exit():
         The Method would then still be executed but can not access attributes of logger anymore.
         Thus, we need to use globals anyway.
     """
+
+    print('_on_exit 1')
+    logger._event_stop.set()
+    logger._q.join()
+    logger._th_print.join()
+    print('_on_exit 2')
+
     # if exception is None:
     t_elapsed = time.time() - t_start
     str_timer_info = ('\n\nTotal runtime: {} ' + LIGHT_GRAY + '(since logbox import)' + RESET + '\n').format(format_time(t_elapsed))
-    sys.stdout.write(str_timer_info)
+    _print(str_timer_info)
 
     # Explicitly log time to all channels
     for channel in global_channels:
         channel.on_log(str_timer_info + ' ' + format_time(time.time()-t_start), global_default_channel)
     
     # Deactivate ANSI escape codes if necessary
-    # sys.stdout.write(RESET)
+    # _print(RESET)
     # time.sleep(0.1)
     deactivate_esc()
+'''
 
 
 # Logger is started here to handle output to the terminal and can be used by functions below
 logger = Logger()
 
 # Exceptions should always be given to the logger
-sys.excepthook = logger._on_exception
-atexit.register(_on_exit)
+sys.excepthook = logger.on_exception
+
+# When program exits (with or without exception) the logger should be informed
+atexit.register(logger.on_exit)
 
 
 def _info(*str_out, channel=DEFAULT_CHANNEL):
