@@ -33,6 +33,9 @@ class ProgbarBase:
         self.str_columns_headline = ''
         self.str_progress = ''
 
+        # Flag will be set by iterator after StopIteration exception
+        self._is_complete = False
+
         self.t_start = None
         self.t_stop = None
 
@@ -56,9 +59,16 @@ class ProgbarBase:
     def stop(self):
         self._on_stop()
 
+    def is_complete(self):
+        return self._is_complete
+
+    def set_complete(self):
+        self._is_complete = True
+
     def _on_start(self):
         self._reset()
         self.t_start = time.time()
+        self._is_complete = False
         self._initial_update()
         self.state = PROGBAR_ACTIVE
 
@@ -170,10 +180,10 @@ class ProgbarKnownLength(ProgbarBase):
         return str_progress
 
     def _parse_progbar(self, i):
-        ratio = i / self.limit_updates
+        ratio = i / self.limit_updates if self.limit_updates != 0 else 1
         n_steps_done = int(round(ratio * self.steps_progbar))
 
-        ratio_parsed = np.round(100 * i / self.limit_updates, decimals=self.n_decimals)
+        ratio_parsed = np.round(100 * ratio, decimals=self.n_decimals)
         str_percent = self.str_percent_template.format(ratio_parsed)
 
         n_chars_left = int(np.ceil((self.steps_progbar - len(str_percent)) / 2))
